 //<>//
/***********************************************************************************\
 .                                                                                  .
 .                                   S P R I T E S                                  .
 .                                                                                  .
 .----------------------------------------------------------------------------------.
 .                                                                                  .
 . Clase que implementa el dibujado de Sprites, para ello se debe suministrar una   .
 . imagen con todos los frames que debe pintar para cada personaje.                 .
 .                                                                                  .
 . Incorpora métodos para realizar movimientos horizontales automáticos, así como   .
 . la detección de colisiones.
 .                                                                                  .
 .    Todas las imagenes y el sonido son de dominio público, llicència CC-0         .
 .                                                                                  .
 -----------------------------------------------------------------------------------.
 .                                                                                  .
 .                    CC-BY -  P E D R O   P U E R T A S                            .
 .                                                                                  .
 .                     MOSAIC 2017 - VIDEOJUEGOS CON PROCESSING                     .
 .                                                                                  .
 \**********************************************************************************/


class Sprite {
  // Variable donde se almacenan todos los frames del sprite
  private PImage img;

  private float x;          // Posición horizontal 
  private float y;          // Posición vertical
  private int wi;           // Ancho 
  private int he;           // Alto
  private int minX=0;       // Coordenada mínima horizontal hasta donde se moverá
  private int maxX=width;   // Coordenada màxima horizontal hasta donde se moverá

  private float speed;      // Indica la velocidad de cambio de frames
  private int nFramesX;     // Guarda el número de frames en horizontal
  private int nFramesY;     // Guarda el número de frames en vertical
  private float nFrame=0;   // Frame actual de la animación

  private int sentit = 1;   // Indica el sentido del moviment
  private int flip=1;       // Indica si se ha de pintar volteado

  private boolean visible = true;

  private Sprite() {
  }

  /**
   Constructor de la clase
   
   @param x    : posición horizontal
   @param y    : posición vertical
   @param nx   : nº de frames horizontales
   @param ny   : nº de frames verticales
   @param file : fitxer
   */
  Sprite(int x, int y, int nx, int ny, String file) {
    img = loadImage(file);
    this.x = x;
    this.y = y;
    this.nFramesX = nx;
    this.nFramesY = ny;
    this.wi = img.width/nx;
    this.he = img.height/ny;
    // Adaptamos la velocidad del cambio entre frames dependiendo del frameRate, esto es porque habrá
    // sprites con muchos frames de moviment y otros con sólo dos. Así nos aseguramos que el movimiento
    // vea fluido, aunque se puede cambiar con la función velocidad()
    speed = (nFramesX*nFramesY)/frameRate;
  }

  /**
   Función que resetea los valores de volteado, número de frame y visibilidad, etc.
  */
  void reset() {
    flip = 1;
    sentit = 1;
    nFrame = 0;
    visible = true;
  }

  /**
   Función que posiciona el sprite en las coordenadas horizontal y vertical indicadas
   
   @param xx - Posición horizontal
   @param yy - Posición vertical
   */
  void pos(int xx, int yy) {
    this.x = xx;
    this.y = yy;
  }

  /**
   Función que fija la velocidad de cambio de frame
   
   @param s - Velocidad de cambio de frame
   */
  void velocidad(float s) {
    this.speed = s;
  }

  /**
   Función que modifica la medida del sprite
   
   @param w - Indica el ancho
   @param h - Indica el alto
   */
  void medida(int w, int h) {
    this.wi = w;
    this.he = h;
  }

  /**
   Función que indica los límites del movimiento horizontal del sprite
   
   @param min - Coordenada horizontal mínima
   @param max - Coordenada horizontal máxima
   */
  void limites(int min, int max) {
    minX = min;
    maxX = max;
    
    if (x < minX) x = minX;
    if (x > maxX) x = maxX;
  }

  /**
   Función que mueve el personaje hacia la derecha haciendo ondas
   
   @param nx - Indica cuantos píxels en horizontal se mueve
   @param ny - Indica en vertical la altura de la onda (si cero movimiento plano)
   */
  void mover_der(float nx, float ny) {
    x += nx;  // Aumentamos la coordenada X
    y += cos(x)*ny;  // Aumentamos Y con el valor del coseno, lo que dará valores positivos y negativos
    if (x > maxX) x = minX; // Si hemos llegado al final volveremos al principio
  }

  /**
   Función que mueve el personaje hacia la izquierda haciendo ondas
   
   @param nx - Indica cuantos píxels en horizontal se mueve
   @param ny - Indica en vertical la altura de la onda (si cero movimiento plano)
   */
  void mover_izq(float nx, float ny) {
    x-=nx;  // Disminuimos la coordenada X
    y += cos(x)*ny; // Aumentamos Y con el valor del coseno, lo que dará valores positivos y negativos
    if (x < minX) x = maxX; // Si hemos llegado al principio volveremos al final
  }

  /**
   Funció que mueve el sprite de izquierda a derecha, gira y vuelve a la izquierda en una nueva posición en altura
   
   @param nx - Indica cuantos píxels en horizontal se mueve
   @param ny - Indica en vertical la altura de la onda (si cero movimiento plano)
   */
  void mover(float xx, float yy) {
    x += (xx*sentit);  // Aumentamos la coordenada X en el sentido que toque (1 -> derecha, -1 -> izquierda)
    y += cos(x)*yy;    // Movimiento ondulatorio vertical

    // Si llegamos a un límite, hemos de voltear el personaje y cambiar la posición vertical si así se ha indicado
    if (x + (xx*sentit) < minX || x + (xx*sentit) > maxX) {
      sentit = -sentit;
      if (yy > 0) y = random(height/2);
      voltea();
    }
  }

  /**
   Función que voltea el sprite
   */
  void voltea() {
    flip = (flip > 0) ? -1 : 1;
  }

  /**
   Función que indica si se ha de pintar el sprite volteado o no
   */
  boolean volteado() {
    return flip < 0;
  }

  /**
   Función que devuelve el número de frame a pintar actual
  */
  int getFrame() {
    return (int)nFrame;
  }

  /**
   Función que devuelve la coordenada X donde está posicionado el sprite
  */
  int getX() {
    return (int)x;
  }
  
  /**
   Función que devuelve la coordenada Y donde está posicionado el sprite
  */
  int getY() {
    return (int)y;
  }
  /**
   Funció de pintado en la que cada vez que es llamada se pinta el siguiente frame (animación)
   */
  void draw() {
    draw(-1);
  }

  /**
   Función de pintado, se le puede indicar el frame exacto o -1 si queremos que haga la
   animación
   
   @param nf - Número de frame a pintar. Si -1 pinta animación
   */
  void draw(int nf) {
    
    if (!visible) return;
    
    // Si el número de frame es negativo, cambiamos de frame a la velocidad fijada
    if (nf < 0) {
      nFrame += speed;
      // Si llegamos al último frame volvemos al primero
      if (nFrame >= nFramesX*nFramesY) nFrame = 0;
    } else {
      // Fijamos el frame 
      nFrame = nf;
    }
    // Calculamos el ancho i alto del sprite a partir del número de frames en X e Y
    int w = img.width/nFramesX;
    int h = img.height/nFramesY;

    // Crea una imagen temporal para pintar del ancho y alto declarado
    PImage frame = createImage(this.wi, this.he, ARGB);

    // Calculamos el trozo de imagen que hemos de tomar desde la imagen completa y se escala al ancho y alto declarado. 
    frame.copy(img, (int)(nFrame%nFramesX)*w, (int)(nFrame/nFramesX)*h, w, h, 0, 0, this.wi, this.he);

    // Como la función scale afecta al sistema de coordenades de todo el escenario, primero hemos de guardar su estado con la instrucción pushMatrix 
    // despues lo escalamos, pintamos y volvemos a recuperar el estado guardado (popMatrix)
    pushMatrix();            // Guardamos el estado actual
    scale(flip, 1.0);        // Escalamos a flip (si es -1 hace un reflejo horizontal, el vertical no se modifica
    image(frame, x*flip, y); // Si lo hemos volteado, les coordenades van justo al revés, es decir el punto 0 està a la derecha, lo controlamos multiplicando X por la variable flip
    popMatrix();             // Una vez pintado en la pantalla, recuperamos el estado del sistema para que no afecte al resto de objetos
  }

  /**
   Indica si se ha de pintar o no el sprite
  */
  void visible(boolean b) {
    visible = b;
  }
 
  /**
   Función que mira si ha colisionado con otro sprite
  */
  boolean colisionan(Sprite b) {
    // Calculamos el centre del sprite a comparar
    int bx = (int)b.x+b.wi*b.flip/2;
    
    // Calculamos el centro del sprite
    int mx = (int)x+wi/2*flip;
    int my = (int)y+he/2;
    
    // Si el centro està dentro del recuadro del sprite comparado es que han colisionado
    return (b.visible && mx >= bx-b.wi/2 && mx <= bx+b.wi/2 && my >= b.y && my <= b.y+b.he); 
  }
 
  /**
   Función que mira si un punto está dentro del sprite
  */
  boolean dentro(float xx, float yy) {
    // Calculamos el centro del sprite
    int mx = (int)x+wi/2*flip; // Hay que tener en cuenta si está volteado
    int my = (int)y+he/2;
    
    // Si el punto está dentro del rectángulo devolverá true
    return (xx > mx-wi/2 && xx < mx+wi/2 && yy > my-he/2 && yy < my+he/2);
  }
 
  /**
    Crea una copia del sprite
  */
  public Sprite clonar() {
    Sprite p = new Sprite();
    p.img = this.img;
    p.nFramesX = this.nFramesX;
    p.nFramesY = this.nFramesY;
    p.nFrame = random(nFramesX*nFramesY);
    return p;
  }  
}